import json
import os
import uuid
from termcolor import colored
from websockets import Headers
import yaml
from websockets.sync.client import connect

from gru import DEFAULT_CANSO_API_CONFIG
from gru.agents.models import (
    MemoryInsertRequest,
    MemoryUpdateRequest,
    MemoryDeleteRequest,
)
from gru.agents.models import (
    AgentPromptRequest,
    AgentRegisterRequest,
    AgentUpdateRequest,
)
from gru.schema.api_request_handler import APIRequestHandler
from gru.schema.api_response_handler import APIResponseHandler
from gru.utils.config_reader import ConfigReader

from cookiecutter.main import cookiecutter

AGENT_CONFIG_FILE_NAME = "config.yaml"

AI_AGENT_TEMPLATE_URL = "https://github.com/Yugen-ai/canso-ai-agent-templates.git"


def ai_agent_templates_setup():
    """
    Run Cookiecutter with the specified template.
    """
    cookiecutter(AI_AGENT_TEMPLATE_URL)


def call_api(
    correlation_id: str,
    auth_token: str,
    base_url: str,
    endpoint: str,
    method: str,
    body: dict | None = None,
):
    request_handler = APIRequestHandler(auth_token, base_url=base_url)
    request_handler.set_correlation_id(correlation_id)

    response = request_handler.send_request(method, endpoint, body)

    response_handler = APIResponseHandler(response)
    response_handler.check_for_errors()
    response_message = response_handler.get_message()

    return response_message


def register_agent(
    correlation_id: str,
    auth_token: str,
    agent_folder: str,
    cluster_name: str,
    image: str,
    image_pull_secret: str,
):
    config_file_path = os.path.join(agent_folder, AGENT_CONFIG_FILE_NAME)

    with open(config_file_path, "r") as config_file:
        config_dict = yaml.safe_load(config_file)

    agent_slug = config_dict["agent_name"].lower().replace(" ", "-")

    agent_register_request = AgentRegisterRequest(
        cluster_name=cluster_name,
        agent_name=agent_slug,
        image=image,
        image_pull_secret=image_pull_secret,
        task_server_name=config_dict["task_server_name"],
        checkpoint_db_name=config_dict["checkpoint_db_name"],
        replicas=config_dict["replicas"],
        iam_role_arn=config_dict["iam_role_arn"],
        vector_db_name=config_dict["vector_db_name"],
    )

    body = agent_register_request.model_dump()

    configs = ConfigReader(DEFAULT_CANSO_API_CONFIG)
    response_message = call_api(
        correlation_id,
        auth_token,
        configs.agents_service_url,
        configs.agent_register_endpoint,
        "post",
        body,
    )

    return response_message


def update_agent(
    correlation_id: str,
    auth_token: str,
    agent_folder: str,
    image: str,
    image_pull_secret: str,
):
    config_file_path = os.path.join(agent_folder, AGENT_CONFIG_FILE_NAME)

    with open(config_file_path, "r") as config_file:
        config_dict = yaml.safe_load(config_file)

    agent_name = config_dict["agent_name"].lower().replace(" ", "-")

    agent_update_request = AgentUpdateRequest(
        image=image,
        image_pull_secret=image_pull_secret,
        task_server_name=config_dict["task_server_name"],
        checkpoint_db_name=config_dict["checkpoint_db_name"],
        replicas=config_dict["replicas"],
        iam_role_arn=config_dict["iam_role_arn"],
    )

    body = agent_update_request.model_dump()

    configs = ConfigReader(DEFAULT_CANSO_API_CONFIG)
    endpoint = configs.agent_update_endpoint.replace("{agent_name}", agent_name)

    response_message = call_api(
        correlation_id, auth_token, configs.agents_service_url, endpoint, "patch", body
    )

    return response_message


def deploy_agent(correlation_id: str, auth_token: str, agent_name: str):

    configs = ConfigReader(DEFAULT_CANSO_API_CONFIG)
    endpoint = configs.agent_deploy_endpoint.replace("{agent_name}", agent_name)

    response_message = call_api(
        correlation_id, auth_token, configs.agents_service_url, endpoint, "post"
    )

    return response_message


def delete_agent(correlation_id: str, auth_token: str, agent_name):

    configs = ConfigReader(DEFAULT_CANSO_API_CONFIG)
    endpoint = configs.agent_delete_endpoint.replace("{agent_name}", agent_name)

    response_message = call_api(
        correlation_id, auth_token, configs.agents_service_url, endpoint, "delete"
    )

    return response_message


def prompt_agent(agent_name: str, prompt: dict, auth_token: str):
    """
    Send a prompt to a deployed agent.

    Args:
        agent_name (str): Name of the deployed agent
        prompt (dict): Dictionary containing the prompt data
        auth_token (str): Authentication token for API access

    Returns:
        AgentPromptResponse: Response containing the prompt ID

    Raises:
        ApiError: If the API request fails
        ValueError: If the request parameters are invalid
    """
    prompt_request = AgentPromptRequest(prompt=prompt)
    request_json = prompt_request.model_dump()

    configs = ConfigReader(DEFAULT_CANSO_API_CONFIG)
    request_handler = APIRequestHandler(auth_token, configs)

    endpoint = configs.agent_prompt_endpoint.replace("{agent_name}", agent_name)

    response = request_handler.send_request("post", endpoint, request_json)
    response_handler = APIResponseHandler(response)
    response_handler.check_for_errors()

    response_data = response_handler.get_success_data()
    return response_data


def read_prompt_file(prompt_file: str) -> dict:
    """
    Read and parse a JSON prompt file.

    Args:
        prompt_file (str): Path to the JSON file containing prompt data

    Returns:
        dict: Parsed prompt data

    Raises:
        FileNotFoundError: If the file doesn't exist
        json.JSONDecodeError: If the file contains invalid JSON
    """
    try:
        with open(prompt_file, "r") as f:
            return json.load(f)
    except FileNotFoundError:
        raise FileNotFoundError(f"Prompt file not found: {prompt_file}")
    except json.JSONDecodeError as e:
        raise json.JSONDecodeError(
            f"Invalid JSON in {prompt_file}: {str(e)}", e.doc, e.pos
        )


def converse_agent(
    auth_token: str, agent_name: str, conversation_id: str | None = None
):

    headers = Headers({"authorization": auth_token})

    if conversation_id is None:
        conversation_id = str(uuid.uuid4())[:8]

    configs = ConfigReader(DEFAULT_CANSO_API_CONFIG)
    uri = configs.agent_converse_endpoint.replace("{agent_name}", agent_name).replace(
        "{conversation_id}", conversation_id
    )

    with connect(uri, additional_headers=headers) as websocket:
        print(f"Conversation ID: {conversation_id}")
        while True:
            user_msg = input(colored("User: ", attrs=["bold"]))
            if user_msg == "STOP":
                break

            websocket.send(user_msg)
            response = websocket.recv()
            print(f"{colored('Agent: ', attrs=['bold'])}{response}")

    return "Conversation Stopped"


def memory_insert(
    correlation_id: str, auth_token: str, agent_name: str, file_path: str
) -> str:
    """
    Insert data into agent memory
    Args:
        correlation_id: Unique identifier for the request
        auth_token: Authentication token
        agent_name: Name of the agent
        file_path: Path to JSON file containing memory data
    Returns:
        API response message
    """
    try:
        if not os.path.exists(file_path):
            return f"Error: File not found: {file_path}"

        try:
            with open(file_path, "r") as f:
                memory_data = json.load(f)
        except json.JSONDecodeError as e:
            line_col = f"line {e.lineno}, column {e.colno}"
            return f"Error: Invalid JSON in {file_path} at {line_col}: {e.msg}"

        required_fields = ["collection_name", "text", "data"]
        missing_fields = [
            field for field in required_fields if field not in memory_data
        ]
        if missing_fields:
            return (
                f"Error: Missing required fields in JSON: {', '.join(missing_fields)}"
            )

        try:
            memory_insert_request = MemoryInsertRequest(
                collection_name=memory_data["collection_name"],
                text=memory_data["text"],
                data=memory_data.get("data", {}),
            )
        except Exception as e:
            return f"Error: Invalid request data: {str(e)}"

        body = memory_insert_request.model_dump()

        configs = ConfigReader(DEFAULT_CANSO_API_CONFIG)
        endpoint = configs.agent_memory_endpoint.replace("{agent_name}", agent_name)
        agent_url = f"{configs.agents_service_url}/agents/{agent_name}"

        request_handler = APIRequestHandler(auth_token, base_url=agent_url)
        request_handler.set_correlation_id(correlation_id)

        try:
            response = request_handler.send_request("post", endpoint, body)
            response_handler = APIResponseHandler(response)
            response_handler.check_for_errors()
            return response_handler.get_message()
        except Exception as e:
            if hasattr(e, "status_code") and hasattr(e, "message"):
                return f"API Error ({e.status_code}): {e.message}"
            else:
                return f"Error communicating with the API: {str(e)}"

    except Exception as e:
        return f"Unexpected error: {str(e)}. Correlation ID: {correlation_id}"


def memory_update(
    correlation_id: str, auth_token: str, agent_name: str, file_path: str
) -> str:
    """
    Update data in agent memory
    Args:
        correlation_id: Unique identifier for the request
        auth_token: Authentication token
        agent_name: Name of the agent
        file_path: Path to JSON file containing update data
    Returns:
        API response message
    """
    try:
        if not os.path.exists(file_path):
            return f"Error: File not found: {file_path}"

        try:
            with open(file_path, "r") as f:
                memory_data = json.load(f)
        except json.JSONDecodeError as e:
            line_col = f"line {e.lineno}, column {e.colno}"
            return f"Error: Invalid JSON in {file_path} at {line_col}: {e.msg}"

        required_fields = ["collection_name", "memory_id", "text", "data"]
        missing_fields = [
            field for field in required_fields if field not in memory_data
        ]
        if missing_fields:
            return (
                f"Error: Missing required fields in JSON: {', '.join(missing_fields)}"
            )

        memory_id = memory_data.get("memory_id", "")
        if not memory_id or not isinstance(memory_id, str):
            return f"Error: Invalid memory_id: must be a non-empty string"

        try:
            memory_update_request = MemoryUpdateRequest(
                collection_name=memory_data["collection_name"],
                memory_id=memory_data["memory_id"],
                text=memory_data["text"],
                data=memory_data.get("data", {}),
            )
        except Exception as e:
            return f"Error: Invalid request data: {str(e)}"

        body = memory_update_request.model_dump()

        configs = ConfigReader(DEFAULT_CANSO_API_CONFIG)
        endpoint = configs.agent_memory_endpoint.replace("{agent_name}", agent_name)
        base_url = f"{configs.agents_service_url}/agents/{agent_name}"

        request_handler = APIRequestHandler(auth_token, base_url=base_url)
        request_handler.set_correlation_id(correlation_id)

        try:
            response = request_handler.send_request("patch", endpoint, body)
            response_handler = APIResponseHandler(response)
            response_handler.check_for_errors()
            return response_handler.get_message()
        except Exception as e:
            if hasattr(e, "status_code") and hasattr(e, "message"):
                return f"API Error ({e.status_code}): {e.message}"
            else:
                return f"Error communicating with the API: {str(e)}"

    except Exception as e:
        return f"Unexpected error: {str(e)}. Correlation ID: {correlation_id}"


def memory_delete(
    correlation_id: str,
    auth_token: str,
    agent_name: str,
    memory_id: str,
    collection: str,
) -> str:
    """
    Delete data from agent memory
    Args:
        correlation_id: Unique identifier for the request
        auth_token: Authentication token
        agent_name: Name of the agent
        memory_id: ID of the memory to delete
        collection: Name of the collection
    Returns:
        API response message
    """
    try:
        if not memory_id:
            return "Error: memory_id is required and cannot be empty"

        if not collection:
            return "Error: collection name is required and cannot be empty"

        try:
            memory_delete_request = MemoryDeleteRequest(
                collection_name=collection, memory_id=memory_id
            )
        except Exception as e:
            return f"Error: Invalid request data: {str(e)}"

        body = memory_delete_request.model_dump()

        configs = ConfigReader(DEFAULT_CANSO_API_CONFIG)
        endpoint = configs.agent_memory_endpoint.replace("{agent_name}", agent_name)

        base_url = f"{configs.agents_service_url}/agents/{agent_name}"

        request_handler = APIRequestHandler(auth_token, base_url=base_url)
        request_handler.set_correlation_id(correlation_id)

        try:
            response = request_handler.send_request("delete", endpoint, body)
            response_handler = APIResponseHandler(response)
            response_handler.check_for_errors()
            return response_handler.get_message()
        except Exception as e:
            if hasattr(e, "status_code") and hasattr(e, "message"):
                return f"API Error ({e.status_code}): {e.message}"
            else:
                return f"Error communicating with the API: {str(e)}"

    except Exception as e:
        return f"Unexpected error: {str(e)}. Correlation ID: {correlation_id}"



from datetime import datetime


def dt_to_str(dt_obj: datetime):
    """
    Takes a datetime object and converts it to a STRING having a set datetime format (%Y-%m-%d %H:%M:%S)

    Parameters
    ----------
    dt_obj : A datetime object

    Returns
    -------
    str : A string representation of dt_obj
    """
    modified_date = dt_obj.strftime("%Y-%m-%d %H:%M:%S")
    modified_date = datetime.strptime(modified_date, "%Y-%m-%d %H:%M:%S")

    return modified_date


def update_end_date(start_date: datetime, dry_run_duration: datetime):
    """
    Takes start_date and dry_run_duration as an input to this function and updates the  end date  based on deafult_dry_run_duration_value

    Parameters
    ----------
    start_date : A datetime object
    dry_run_duration: A datetime object

    Returns
    -------
    datetime Object: Returns updated end date.

    """
    end_date = start_date + dry_run_duration
    return end_date


def is_duration_valid_for_dry_run(start_date: datetime, end_date: datetime, max_dry_run_duration_days: int):
    """
    Takes start_date and end_date as an input to this function and validates it based on different conditions.

    Parameters
    ----------
    start_date : A datetime object
    end_date   : A datetime object

    Returns
    -------
    bool : returns boolean value based on condition checks.
    """
    if start_date is None:
        raise ValueError("start_date cannot be None")
    elif end_date is None:
        return True
    else:
        if end_date < start_date:
            raise ValueError("end_date should be greater than start_date")
    start_date = dt_to_str(start_date)
    end_date = dt_to_str(end_date)
    date_delta = end_date - start_date
    num_of_days = date_delta.days
    return num_of_days <= max_dry_run_duration_days



from enum import Enum
from typing import Optional
from gru.agents.tools.core.embeddings.openai import OpenAIEmbeddingGenerator
# from gru.agents.tools.core.embeddings.sentence_transformer import SentenceTransformerEmbeddingGenerator
from gru.agents.tools.core.embeddings.base import EmbeddingGenerator

class EmbeddingType(Enum):
    """Enum for supported embedding types"""
    OPENAI = "openai"
    # SENTENCE_TRANSFORMER = "sentence_transformer"

class EmbeddingFactory:
    """Factory class for creating embedding generators"""
    @staticmethod
    def get_embedding_generator(
        embedding_type: EmbeddingType,
        client = None,
        model: Optional[str] = None,
        batch_size: Optional[int] = None
        ) -> EmbeddingGenerator:
        """Get an embedding generator based on the type"""
        if embedding_type == EmbeddingType.OPENAI:
            return OpenAIEmbeddingGenerator(client=client, model=model, batch_size=batch_size)
        # elif embedding_type == EmbeddingType.SENTENCE_TRANSFORMER:
        #     if model is None:
        #         model = "all-MiniLM-L6-v2"
        #     if batch_size is None:
        #         batch_size = 32
        #     return SentenceTransformerEmbeddingGenerator(client=client, model=model, batch_size=batch_size)
        else:
            raise ValueError(f"Unsupported embedding type: {embedding_type}")


import requests
from typing import Union, Optional


class APIResponseHandler:
    """
    Handles responses from the Yugen remote API.

    Attributes:
        response: The raw response from the API.
    """

    def __init__(self, response: requests.Response):
        """
        Initialize the ResponseHandler with a response object.

        Args:
            response: The requests.Response object to handle.
        """
        self.response = response
        self.status_code = self.response.status_code

    def check_for_errors(self):
        """
        Checks if the response contains an error and raises appropriate exceptions.
        """
        response_json = self.response.json()

        if "error" in response_json:
            error = response_json["error"]
            if self.status_code == 500:
                raise InternalServerError(
                    status_code=self.status_code,
                    title="Server Error",
                    message=error["message"],
                )
            elif self.status_code == 400:
                raise BadRequestError(
                    status_code=self.status_code,
                    title="Bad Request",
                    message=error["message"],
                )
            elif self.status_code == 404:
                raise NotFoundError(
                    status_code=self.status_code,
                    title="Entity Not Found",
                    message=error["message"],
                )
            else : 
                raise ApiError(
                    status_code=self.status_code,
                    title=error["title"],
                    message=error["message"],       
                )

            

    def get_success_data(self) -> Optional[Union[dict, list]]:
        """
        Returns the parsed data from a successful response.

        Returns:
            The parsed data if a "data" key is found in the response, otherwise None.
        """
        response_json = self.response.json()
        return response_json.get("data") if isinstance(response_json, dict) and "data" in response_json else response_json

    
    def get_message(self) -> str:
        response_json = self.response.json()
        return response_json.get("message")


class ApiError(Exception):
    """
    Represents an error returned by the API.
    """

    def __init__(
        self, status_code: int, title: str, message: str, *args, **kwargs
    ):
        """
        Initializes the ApiError with details from the API response.
        """
        super().__init__(title, message, *args, **kwargs)
        self.status_code = status_code
        self.title = title
        self.message = message

class InternalServerError(ApiError):
    """
    Represents an InternalServerError returned by the API.
    """

    def __init__(
        self, status_code: int, message: str, title: str = "Internal Server Error", *args, **kwargs
    ):
        """
        Initializes the ApiError with details from the API response.
        """
        super().__init__(status_code, title, message,  *args, **kwargs)

class BadRequestError(ApiError):
    """
    Represents a BadRequestError returned by the API.
    """

    def __init__(
        self, status_code: int, message: str, title: str = "Bad Request", *args, **kwargs
    ):
        """
        Initializes the BadRequestError with details from the API response.
        """
        super().__init__(status_code, title, message,  *args, **kwargs)

class NotFoundError(ApiError):

    """
    Represents a NotFoundError returned by the API.
    """

    def __init__(
        self, status_code: int, message: str, title: str = "Entity Not Found", *args, **kwargs
    ):
        """
        Initializes the NotFoundError with details from the API response.
        """
        super().__init__(status_code, title, message,  *args, **kwargs)






"""
In Machine Leaning, Training Data is the Data set on top of which Machine Learning Models are trained.

Training Data is generated by joining historical features created using Create Raw & Derived Feature Modules. Generation of Training Data implements point-in-time joins to ensure there's no data leakage.
Function:
    - print
    - register
    - generate
"""

from gru.utils.utils import check_data_type
from gru.utils.entity_type import EntityType
from typing import Dict, List, Optional
from gru.features.processing_engine_configs import BatchProcessingEngineConfigs
from gru.utils.processing_engine import ProcessingEngine


class TrainingData:
    """
    Training data is the join of historical data and the features for that historical data. It is used to train our Machine Learning Models

    Parameters
    ----------
    self: str
        Python self object
    name: str
        Name of the training data object
    description:
        description of the training data
    historical_data_source: str
        It'll be the name of historical data source registered in our registry. We'll fetch the timestamp field and timestamp format from the registry. So the end DS/DE doesn't have to mention it multiple times.
    entities:
        List of columns to be treated as entities in the historical data
    features:
        List of features to be joined with Historical data
    ttl:
        TTL (in minutes) used to determine how far back point-in-time joins should look
    owners:
        list of owners for the training data
    timeout:
        timeout value set for the job's execution to complete. By default value set to 30 mins.


    Returns
    -------
    None

    Notes
    -----
    Time to register the Training data will be generated in Stuart and then will be written to DB. And same will be passed on to Jerry and Bob

    Examples
    --------
    """

    def __init__(
        self,
        name: str,
        description: str,
        historical_data_source: str,
        entities: List[str],
        features: str,
        ttl: int,
        owners: List[str],
        timeout: int = 30,
        k8s_configs: Dict = {},
        processing_engine: Optional[ProcessingEngine] = ProcessingEngine.PYSPARK_K8S,
        processing_engine_configs: Optional[BatchProcessingEngineConfigs] = None,
    ):
        airflow_configs = {"timeout": timeout}

        self.type = EntityType.TRAINING_DATA.value
        self.entity_type = self.type
        self.ttl = ttl
        self.name = name
        self.owners = owners
        self.k8s_configs = k8s_configs
        self.entities = entities
        self.features = features
        self.description = description
        self.historical_data_source = historical_data_source
        self.airflow_configs = airflow_configs
        self.processing_engine = processing_engine
        self.processing_engine_configs = (
            processing_engine_configs or BatchProcessingEngineConfigs()
        )

        self.field_datatypes_dict = {
            int: [
                self.ttl,
                self.airflow_configs["timeout"],
            ],
            str: [
                self.name,
                self.description,
                self.historical_data_source,
            ],
            list: [
                self.owners,
                self.features,
            ],
        }

        self.args_type_check = check_data_type(self.field_datatypes_dict)

        if not self.args_type_check:
            return

    def to_json(self) -> Dict:
        return {
            "type": self.type,
            "ttl": self.ttl,
            "name": self.name,
            "owners": self.owners,
            "airflow_configs": self.airflow_configs,
            "entities": self.entities,
            "features": self.features,
            "description": self.description,
            "k8s_configs": self.k8s_configs,
            "processing_engine": self.processing_engine.value,
            "processing_engine_configs": self.processing_engine_configs.to_json(),
            "historical_data_source": self.historical_data_source,
        }



import yaml
from gru import DEFAULT_CANSO_API_CONFIG
from gru.components.models import ResourceSetupRequest
from gru.schema.api_request_handler import APIRequestHandler
from gru.schema.api_response_handler import APIResponseHandler
from gru.utils.config_reader import ConfigReader

def setup(correlation_id:str, cluster_name: str, config_file:str, auth_token: str):
    
        with open(config_file, "r") as config_file:
            config_dict = yaml.safe_load(config_file)

        resource_setup_request = ResourceSetupRequest(**config_dict)
        resource_setup_request.cluster_name = cluster_name
        request_json = resource_setup_request.model_dump()

        configs = ConfigReader(DEFAULT_CANSO_API_CONFIG)
        request_handler = APIRequestHandler(auth_token, configs, base_url=configs.agents_service_url)
        request_handler.set_correlation_id(correlation_id)
        response = request_handler.send_request("post", configs.component_setup_endpoint, request_json)
        response_handler = APIResponseHandler(response)

        return response_handler.get_message()


"""
To update existing preprocessing transform. Supported values - 
    - "owners"
    - "description"
    - "start_time"
"""
from typing import Dict
from datetime import datetime


class UpdatePreprocessingTransform:
    """
    Update a preprocessing transform 

    Parameters
    ----------
    transform_name: str
        preprocessing transform name to be updated
    property: str
        name of the property to be updated
    new_val: str
        new value for the transform property
    
    Returns
    -------
    None

    Notes
    -----

    Examples
    --------
    >>> obj = UpdatePreprocessingTransform(
    ...         name= 'pre_process_movie_collection',
    ...         property= 'active',
    ...         new_val= False,
    ... )
    ... _
    ... yugen_client.update(update_obj)
    """

    def __init__(self, transform_name, property, new_val):
        self.name = transform_name
        self.property = property
        self.new_val = new_val

        if isinstance(self.new_val, datetime):
            self.new_val = str(self.new_val)


    def to_json(self) -> Dict:
        """
        Convert the UpdateFeature object to a JSON-compatible dictionary.
        """
        return {
            "transform_name": self.name,
            "property": self.property,
            "new_value": self.new_val,
        }


from typing import List, Any
from gru.agents.tools.core.embeddings.base import EmbeddingGenerator


class OpenAIEmbeddingGenerator(EmbeddingGenerator):
    MODELS = {
        "text-embedding-3-small": 1536,
        "text-embedding-3-large": 3072,
        "text-embedding-ada-002": 1536
    }

    def __init__(
            self,
            client: Any,
            model: str = "text-embedding-3-small",
            batch_size: int = 100
    ):
        self.client = client
        self.model = model or "text-embedding-3-small"
        if self.model not in self.MODELS:
            raise ValueError(f"Model {model} not supported. Choose from {list(self.MODELS.keys())}")
        self.batch_size = batch_size
        self._dimension = self.MODELS[self.model]

    @property
    def dimension(self) -> int:
        return self._dimension


    async def generate(self, text: str) -> List[float]:
        response = self.client.embeddings.create(
            input=text,
            model=self.model
        )
        return response.data[0].embedding


    async def generate_batch(self, texts: List[str]) -> List[List[float]]:
        all_embeddings = []
        for i in range(0, len(texts), self.batch_size):
            batch = texts[i:i + self.batch_size]
            response = self.client.embeddings.create(
                input=batch,
                model=self.model
            )
            embeddings = [data.embedding for data in response.data]
            all_embeddings.extend(embeddings)
        return all_embeddings



"""
ML Server Interface.

This abstract base class defines the interface for a machine learning server.

Methods:
    - start: Abstract method to start the server.
    - stop: Abstract method to stop the server.
    - reload: Abstract method to reload the server.
    - load: Abstract method to load the server.
"""

from abc import ABC, abstractmethod

class MLServerInterface(ABC):

    @abstractmethod
    def start(self):
        """Abstract method to start the server."""
        pass

    @abstractmethod
    def stop(self):
        """Abstract method to stop the server."""
        pass

    @abstractmethod
    def reload(self):
        """Abstract method to reload the server."""
        pass

    @abstractmethod
    def load(self):
        """Abstract method to load the server."""
        pass




from typing import List, Dict, Any, Optional, Tuple
from gru.agents.tools.core.code_generator.models import RetrievalResult
from gru.agents.tools.core.vector_db.milvus import MilvusClient
from .base import ContextRetriever


class SQLContextRetriever(ContextRetriever):
    def __init__(
        self,
        vector_store: MilvusClient,
        system_prompt: Optional[str] = None,
        user_prompt_template: Optional[str] = None,
    ):
        super().__init__(vector_store, system_prompt, user_prompt_template)
        self.vector_store = vector_store

    async def get_relevant_datasources_and_schemas(
        self,
        query_embedding: List[float],
        top_k: int,
        embedding_field: str,
        search_params: Optional[Dict[str, Any]] = None,
        output_fields: Optional[List[str]] = None,
    ) -> Tuple[List[str], List[str]]:
        """
        Retrieves relevant data sources (tables) and their schemas.
        """
        if search_params is None:
            search_params = {"metric_type": "L2", "params": {"nprobe": 10}}

        if output_fields is None:
            output_fields = ["table_name", "contents"]

        # Perform the similarity search
        search_response = await self.vector_store.similarity_search(
            collection_name="canso_table_metadata",
            query_vector=query_embedding,
            anns_field=embedding_field,
            top_k=top_k,
            search_params=search_params,
            output_fields=output_fields,
        )

        # Check if the search was successful
        if search_response.status != "success":
            raise ValueError(f"Search failed: {search_response.message}")

        # Extract the actual results from the search response
        data = search_response.results

        # Extract table names and schemas
        tables = [item.get("table_name", "") for item in data]
        schemas = [item.get("contents", "") for item in data]

        return tables, schemas

    async def get_similar_examples(
        self,
        query_embedding: List[float],
        top_k: int,
        embedding_field: str,
        search_params: Optional[Dict[str, Any]] = None,
        output_fields: Optional[List[str]] = None,
    ) -> List[str]:
        """Retrieves similar examples based on the query embedding."""
        if output_fields is None:
            output_fields = ["examples"]

        search_response = await self.vector_store.similarity_search(
            collection_name="canso_examples",
            query_vector=query_embedding,
            anns_field=embedding_field,
            top_k=top_k,
            search_params=search_params,
            output_fields=output_fields,
        )

        # Check if the search was successful
        if search_response.status != "success":
            raise ValueError(f"Search failed: {search_response.message}")

        # Extract the actual results from the search response
        data = search_response.results

        return [item.get("examples", "") for item in data]

    async def get_documentation(
        self,
        query_embedding: List[float],
        top_k: int,
        embedding_field: str,
        search_params: Optional[Dict[str, Any]] = None,
        output_fields: Optional[List[str]] = None,
    ) -> List[str]:
        """Retrieves relevant documentation based on the query embedding."""
        if output_fields is None:
            output_fields = ["documentation"]

        search_response = await self.vector_store.similarity_search(
            collection_name="canso_domain_knowledge",
            query_vector=query_embedding,
            anns_field=embedding_field,
            top_k=top_k,
            search_params=search_params,
            output_fields=output_fields,
        )

        # Check if the search was successful
        if search_response.status != "success":
            raise ValueError(f"Search failed: {search_response.message}")

        # Extract the actual results from the search response
        data = search_response.results

        return [item.get("documentation", "") for item in data]

    async def retrieve_context(
        self, query_embedding: List[float], top_k: int
    ) -> RetrievalResult:
        """Retrieves all relevant context for the query embedding."""
        tables, schemas = await self.get_relevant_datasources_and_schemas(
            query_embedding, top_k, "embedding"
        )
        examples = await self.get_similar_examples(
            query_embedding, top_k, "embeddings"
        )
        documentation = await self.get_documentation(
            query_embedding, top_k, "embeddings"
        )

        return RetrievalResult(
            tables=tables,
            schemas=schemas,
            documentation=documentation,
            examples=examples,
            low_cardinality_values=[],
            domain_knowledge=[],
            opt_rules=[],
        )



from datetime import datetime
from typing import Any, Dict, List, Literal, Optional
from pydantic import BaseModel


class WorkflowRulesQueryParams(BaseModel):
    """Query parameters for listing workflow rules"""
    stage: Optional[str] = None
    status: Optional[str] = None

class RiskRule(BaseModel):
    """Schema for rule response"""
    rule_name: str
    rule_definition: Dict[str, Any]  # Contains operator and sub_rules
    stage: Literal['REGISTERED', 'BACKTESTING', 'STAGING', 'PRODEXPERIMENT', 'LIVE']
    status: Literal['ACTIVE', 'INACTIVE', 'DEPRECATED']
    created_at: datetime
    updated_at: datetime

class SubRule(BaseModel):
    field: str
    operator: Literal['<', '>', '==', '!=', '>=', '<=']
    value: Optional[Any] = None
    redis_key: Optional[str] = None
    redis_field: Optional[str] = None


class BaseRuleRequest(BaseModel):
    """Base model for rule requests"""
    operator: Literal['AND', 'OR']
    sub_rules: List[SubRule]

class CreateRuleRequest(BaseRuleRequest):
    """Model for creating a new rule"""
    rule_name: str

class UpdateRuleRequest(BaseRuleRequest):
    """Model for updating an existing rule"""
    pass

class UpdateRuleStatusRequest(BaseModel):
    """Request model for updating Rule status"""
    new_status: Literal['ACTIVE', 'INACTIVE', 'DEPRECATED']

class UpdateRuleStageRequest(BaseModel):
    """Request model for updating Rule stage"""
    stage: Literal['REGISTERED', 'BACKTESTING', 'STAGING', 'PRODEXPERIMENT', 'LIVE']

    @classmethod
    def from_str(cls, stage: str) -> 'UpdateRuleStageRequest':
        """
        Create request from string status, with validation
        """
        stage_upper = stage.upper()
        if stage_upper not in ('REGISTERED', 'BACKTESTING', 'STAGING', 'PRODEXPERIMENT', 'LIVE'):
            raise ValueError(
                "Stage must be one of:'REGISTERED', 'BACKTESTING', 'STAGING', 'PRODEXPERIMENT', 'LIVE'"
            )
        return cls(stage=stage_upper) 




"""
Defines a Kafka Data Source
"""

import json
from typing import Any, List, Optional, Union, Mapping, Dict
from datetime import datetime

from gru.sources.datasource import DataSource
from gru.utils.data_processing_paradigms import DataProcessingParadigm
from gru.utils.cloud_providers import CloudProviders


class KafkaSource(DataSource):
    """
    Represents an Apache Kafka Data Source.

    Parameters:
    ----------
    name (str):
        Name of the streaming Kafka data source.
    description (str):
        Description of the Kafka streaming data source.
    bootstrap_server (str):
        IP address or URL of the server hosting the Kafka Data Source.
    topic (Union[List[str], str]):
        Name of the topic(s) from which data will be extracted.
    schema (Dict):
        The schema of the payload in the Kafka topic, using the same format as for batch data sources.
    timestamp_field (str):
        Field in the incoming Kafka data used to individually identify each event.
    timestamp_format (str):
        Format of the timestamp field mentioned above.
    owners (Union[List[str], str]):
        List of owners of the data source.
    read_configs (Mapping[str, any]):
        Additional read configurations.
    cloud_provider (CloudProviders):
        The cloud provider associated with this data source (default is None).
    metadata (Optional[Dict[str, str]]):
        Additional metadata for the data source (default is an empty dictionary).
    additional_read_configs (Optional[Dict[str, Any]]):
        Additional read configurations (default is an empty dictionary).
    created_at (datetime):
        Date and time when the data source was created (default is the current timestamp).


    Notes:
    ------
    This class represents an Kafka Data Source for streaming data processing.

    Examples:
    ---------
    >>> streaming_kafka_source_obj = KafkaSource(
    ...     name="test_kafka_streaming_source_v1.0.4",
    ...     description="random desc of data source",
    ...     owners=["platform@yugen.ai"],
    ...     topic="numtest",
    ...     schema={
    ...         "type": "struct",
    ...         "fields": [
    ...             {"name": "id", "type": "integer", "nullable": True, "metadata": {}},
    ...             {"name": "salary", "type": "integer", "nullable": True, "metadata": {}},
    ...             {"name": "incentive", "type": "integer", "nullable": True, "metadata": {}},
    ...             {
    ...                 "name": "timestamp",
    ...                 "type": "timestamp",
    ...                 "nullable": True,
    ...                 "metadata": {},
    ...             },
    ...         ],
    ...     },
    ...     timestamp_field="timestamp",
    ...     timestamp_format="yyyy-MM-dd HH:mm:ssXXX",
    ...     bootstrap_server="3.110.47.157:9092",
    ...     read_configs={
    ...         "watermark_delay_threshold": 5,
    ...         "starting_timestamp": None,
    ...         "starting_offset_by_timestamp": {},
    ...         "starting_offsets": "latest",
    ...         "fail_on_data_loss": True,
    ...         "include_headers": True,
    ...     },
    ...     cloud_provider=CloudProviders.AWS
    ... )
    """

    def __init__(
        self,
        name: str,
        description: str,
        bootstrap_server: str,
        topic: Union[List[str], str],
        schema: Dict,
        timestamp_field: str,
        timestamp_format: str,
        owners: Union[List[str], str],
        read_configs: Mapping[str, any],
        cloud_provider: CloudProviders = None,
        metadata: Optional[Dict[str, str]] = {},
        additional_read_configs: Optional[Dict[str, Any]] = {},
        created_at=datetime.now(),
    ):
        """
        Constructor of Kafka Class
        """
        super().__init__(
            name=name,
            description=description,
            family="kafka",
            processing_paradigm=DataProcessingParadigm.Streaming,
            cloud_provider=cloud_provider,
            additional_read_configs=additional_read_configs,
            physical_uri=read_configs,
            schema=schema,
            type="KafkaSource",
            event_timestamp_field=timestamp_field,
            event_timestamp_format=timestamp_format,
            metadata=metadata,
            owners=owners,
            created_at=created_at,
        )

        self.physical_uri["topic"] = topic
        self.physical_uri["kafka_bootstrap_servers"] = bootstrap_server

    def to_json(self) -> Dict:
        """
        Convert the KafkaSource instance to a JSON-compatible dictionary.

        Returns:
        Dict: A dictionary representing the KafkaSource.
        """
        return {
            "name": self.name,
            "description": self.description,
            "family": self.family,
            "processing_paradigm": self.processing_paradigm.value,
            "cloud_provider": self.cloud_provider.value
            if self.cloud_provider
            else None,
            "additional_read_configs": self.additional_read_configs,
            "physical_uri": self.physical_uri,
            "file_type": self.file_type.type if self.file_type else "",
            "file_type_properties": json.dumps(self.file_type.options)
            if self.file_type
            else {},
            "schema": self.schema,
            "type": self.type,
            "event_timestamp_field": self.event_timestamp_field,
            "event_timestamp_format": self.event_timestamp_format,
            "metadata": self.metadata,
            "owners": self.owners,
            "created_at": str(self.created_at),
        }



from gru.features.processing_engine_configs import StreamingProcessingEngineConfigs
from gru.features.aggregations import SlidingWindowAggregation
from gru.features.streaming_feature import StreamingFeature
from gru.utils.processing_engine import ProcessingEngine
from gru.utils.read_options import KafkaReadOptions
from gru.utils.data_types import DataType
from gru.features.raw_feature import (
    FeatureLogic,
)
import pytest, os


@pytest.fixture
def streaming_feature_fixture():
    current_dir = os.path.dirname(os.path.abspath(__file__))
    configs_path = os.path.join(
        current_dir, "data", "test_default_processing_engine_configs_streaming.yaml"
    )
    streaming_feature_obj = StreamingFeature(
        name="foo",
        description="bar",
        data_type=DataType.FLOAT,
        data_sources=["baz"],
        read_option_configs=[
            {
                "name": "baz",
                "options": KafkaReadOptions(
                    watermark_delay_threshold="10 seconds",
                    processing_time="120 seconds",
                    starting_offsets="earliest",
                    include_headers=True,
                ).to_json(),
            },
        ],
        staging_sink=["quux"],
        staging_sink_write_option_configs={
            "quux": {
                "output_mode": "append",
                "processing_time": "120 seconds",
                "output_partitions": 20,
            }
        },
        online_sink=["corge"],
        owners=["grault@yugen.ai"],
        feature_logic=FeatureLogic(
            field=["garply"],
            time_window="waldo",
            timestamp_field="fred",
            timestamp_format="yyyy-MM-dd HH:mm:ssXXX",
            transform=SlidingWindowAggregation(
                function="plugh",
                partition_by="xyzzy",
                order_by="thud",
                window_period="spam",
                window_duration="eggs",
            ),
            groupby_keys=["ham", "zork"],
        ),
        processing_engine=ProcessingEngine.PYSPARK_K8S,
        processing_engine_configs=StreamingProcessingEngineConfigs(config_path=configs_path),
        online=True,
        offline=True,
        creation_time="2024-05-21 16:32:19.365631",
        start_time="2024-05-21 16:32:46.711044",
    )
    return streaming_feature_obj


def test_to_json(streaming_feature_fixture):
    expected_result = {
        "name": "foo",
        "type": "streaming_feature",
        "data_type": "float",
        "data_sources": ["baz"],
        "owners": ["grault@yugen.ai"],
        "description": "bar",
        "start_time": "2024-05-21 16:32:46.711044",
        "online": True,
        "online_sink_write_option_configs": None,
        "offline": True,
        "entity": [],
        "feature_logic": {
            "name": "",
            "transform": {
                "function": "plugh",
                "partition_by": "xyzzy",
                "order_by": "thud",
                "window_period": "spam",
                "window_duration": "eggs",
                "function_type": "SlidingWindowAggregation",
            },
            "time_window": "waldo",
            "groupby_keys": ["ham", "zork"],
            "field": ["garply"],
            "filter": None,
            "timestamp_field": "fred",
            "timestamp_format": "yyyy-MM-dd HH:mm:ssXXX",
            "is_feature_logic_custom": False,
        },
        "k8s_configs": {},
        "processing_engine": "pyspark_k8s",
        "processing_engine_configs": {
            "deploy-mode": "foo",
            "spark.kubernetes.driver.memory": "bar",
            "spark.kubernetes.executor.memory": "baz",
            "spark.kubernetes.authenticate.driver.serviceAccountName": "qux",
            "spark.hadoop.fs.s3a.aws.credentials.provider": "quux",
            "spark.kubernetes.container.image.pullSecrets": "corge",
            "spark.kubernetes.file.upload.path": "/opt/spark/work-dir",
            "spark.kubernetes.namespace": "grault",
            "num-executors": 1,
            "spark.executor.cores": 1,
            "spark.driver.cores": 1,
            "spark.kubernetes.driver.limit.cores": "garply",
            "spark.kubernetes.executor.limit.cores": "waldo",
            "spark.executor.instances": 1,
        },
        "sink": {"staging": ["quux"], "online": ["corge"]},
        "creation_time": "2024-05-21 16:32:19.365631",
        "airflow_configs": {},
        "created_at": "2024-05-21 16:32:19.365631",
        "metadata": {},
        "read_option_configs": [
            {
                "name": "baz",
                "options": {
                    "delimiter": "|",
                    "fail_on_data_loss": False,
                    "include_headers": True,
                    "infer_schema": True,
                    "processing_time": "120 seconds",
                    "starting_offsets": "earliest",
                    "starting_offsets_by_timestamp": None,
                    "starting_timestamp": None,
                    "watermark_delay_threshold": "10 " "seconds",
                },
            }
        ],
        "staging_sink_write_option_configs": {
            "quux": {
                "output_mode": "append",
                "processing_time": "120 seconds",
                "output_partitions": 20,
            }
        },
    }

    assert streaming_feature_fixture.to_json() == expected_result



import os
from typing import Any, Optional, Type, Union, cast
from langchain_core.tools import BaseTool
from langchain_core.callbacks import CallbackManagerForToolRun
from pydantic import BaseModel, Field
from langgraph.errors import NodeInterrupt
from langchain_core.runnables import RunnableConfig
from langchain_core.messages.tool import ToolCall

from gru.agents.task_server.celery import TaskMetadata, submit_task


class CansoSQLRunnerToolInput(BaseModel):
    query: str = Field(description="sql query to execute")


class CansoSQLRunnerTool(BaseTool):
    name: str = "run_sql_query"
    description: str = "Used to run sql queries on the database"
    args_schema: Type[BaseModel] = CansoSQLRunnerToolInput
    return_direct: bool = True

    class Config:
        extra = "allow"

    # Note: This is still a crude implementation, Needs improvement

    def __init__(self, db_host, db_port, db_username, db_password, db_name):
        super().__init__(
            db_host=db_host,
            db_port=db_port,
            db_username=db_username,
            db_password=db_password,
            db_name=db_name,
            tool_call_id="",
        )
        self.db_host = db_host
        self.db_port = db_port
        self.db_username = db_username
        self.db_password = db_password
        self.db_name = db_name
        self.tool_call_id = ""

    async def ainvoke(
        self,
        input: Union[str, dict, ToolCall],
        config: Optional[RunnableConfig] = None,
        **kwargs: Any,
    ) -> Any:
        if self._is_tool_call(input):
            tool_call_id: Optional[str] = cast(ToolCall, input)["id"]
            self.tool_call_id = tool_call_id
        return await super().ainvoke(input, config, **kwargs)

    def _run(
        self,
        query: str,
        config: RunnableConfig,
        run_manager: Optional[CallbackManagerForToolRun] = None,
    ) -> str:

        agent_url = os.getenv("AGENT_URL")
        agent_name = os.getenv("AGENT_NAME")

        metadata = TaskMetadata(
            prompt_id=config["metadata"]["thread_id"],
            tool_call_id=self.tool_call_id,
            agent_name=agent_name,
            agent_callback_url=f"http://{agent_url}/task-complete",
        )

        task_attributes = {
            "db_host": self.db_host,
            "db_port": self.db_port,
            "db_username": self.db_username,
            "db_password": self.db_password,
            "db_name": self.db_name,
            "query": query,
        }

        submit_task(self.name, metadata, task_attributes)
        raise NodeInterrupt(f"SQL Query submitted to task server")

    def _is_tool_call(self, x: Any) -> bool:
        return isinstance(x, dict) and x.get("type") == "tool_call"



"""
Derived Features are transformations on existing registered features. A few use-cases could be -

Binarizing a spend feature to create a new is_spend_high feature, which assumes values 0 and 1, depending on a threshold
Binning a continuous feature to create custom buckets, e.g. age_bucket or spend_level based on age or spend
OneHotEncoding
LabelEncoding
... and more
Function:
print_feature
register_feature
deploy_feature
"""
from datetime import datetime
from typing import Any, Dict, List, Optional
from gru.features.feature_logic import FeatureLogic
from gru.utils.data_types import DataType
from gru.features.ml_feature import MLFeature
from gru.utils.entity_type import EntityType
from gru.utils.processing_engine import ProcessingEngine
from gru.features.processing_engine_configs import BatchProcessingEngineConfigs


class DerivedFeature(MLFeature):
    """
    Derived features are transformations on top of one or more existing feature.

    Parameters
    ----------
    name (str):
        The name of the feature.
    data_type (DataType):
        Data type of the feature.
    transform (Dict):
        Aggregation function that returns a Python dictionary from Aggregations Module.
    sink_name (List[str]):
        The name of the destination where the feature will be stored.
    offline (bool):
        If True, the feature will be persisted to the Offline Feature Store. Default is True.
    schedule (str):
        Cron schedule for the feature generation. Ref: https://crontab.guru/
    entity (List[str]):
        List of entities associated with the feature.
    feature_logic (FeatureLogic):
        FeatureLogic object defining the logic of the feature.
    online (bool):
        If True, the feature will be persisted to the Online Feature Store. Default is False.
    description (str):
        Description of the feature definition.
    owners (List[str]):
        List of owner names.
    active (bool):
        If True, the feature will start running as soon as it's deployed. Default is True.
    timeout (str):
        Maximum time for feature calculation before sending an alert. (Not supported as part of MVP)
    k8s_configs (Dict):
        Kubernetes configurations for feature processing.
    processing_engine (str):
        processing engine, Currently platform supports pyspark.
    processing_engine_configs (Dict):
        Processing engine configurations specific to the defined processing engine.
    start_time (datetime):
        Start time of the feature. Default is the current datetime.
    creation_time (datetime):
        Timestamp of feature creation. Default is the current datetime.


    Returns
    -------
    None

    Notes
    -----

    Examples
    --------
    >>> derived_feature = DerivedFeature(
    ...     name="test_derived_14DEC_v11",
    ...     description="Total purchase amount for the store",
    ...     staging_sink=["s3_sink_ml_yugen_internal"],
    ...     online_sink=["elasticache-redis-yugen"],
    ...     data_type=DataType.FLOAT,
    ...     owners=["rutvik.shah@yugen.ai"],
    ...     schedule="0 0 * * *",
    ...     entity=["CASE WHEN cpi> 0.5 THEN 10 ELSE 0 END"],
    ...     processing_engine=ProcessingEngine.PYSPARK_K8S,
    ...     processing_engine_configs=BatchProcessingEngineConfigs(),
    ...     online=False,
    ...     offline=True,
    ...     transform=multiply(
    ...         "raw_to_1_derived_13DEC_v123", "raw_to_2_derived_13DEC_v123"
    ...     ),
    ...     start_time=datetime(2022, 8, 26, 0, 0, 0),
    ... )
    """

    def __init__(
        self,
        name: str,
        data_type: DataType,
        transform: Dict,
        offline: bool,
        staging_sink: List[str],
        schedule: str,
        entity: List[str],
        start_time: datetime,
        feature_logic: FeatureLogic = None,
        online: bool = False,
        online_sink: List[str] = [],
        description: str = "",
        owners: List[str] = [],
        active: bool = False,
        timeout: str = "5m",
        k8s_configs: Dict = {},
        read_option_configs: Optional[List[Dict[str, Any]]] = None,
        staging_sink_write_option_configs: Optional[Dict[str, Any]] = None,
        online_sink_write_option_configs: Optional[Dict[str, Any]] = None,
        processing_engine: Optional[ProcessingEngine] = ProcessingEngine.PYSPARK_K8S,
        processing_engine_configs: Optional[BatchProcessingEngineConfigs] = None,
        creation_time: datetime = datetime.now(),
        tags: Dict[str, Any] = {}
    ):
        data_sources = transform["feature_list"]

        feature_logic = {
            "transformation_function": transform["transform"],
            "transform_args": self._get_feature_transform_args(transform),
            "groupby_keys": transform["groupby_keys"]
        }
        airflow_configs = {"active": active, "timeout": timeout, "schedule": schedule}
        sink = {"staging": staging_sink, "online": online_sink}

        super().__init__(
            name=name,
            type=EntityType.DERIVED_FEATURE.value,
            data_type=data_type,
            online=online,
            offline=offline,
            entity=entity,
            data_sources=data_sources,
            sink=sink,
            feature_logic=feature_logic,
            description=description,
            owners=owners,
            start_time=start_time,
            k8s_configs=k8s_configs,
            processing_engine=processing_engine,
            processing_engine_configs=processing_engine_configs
            or BatchProcessingEngineConfigs(),
            airflow_configs=airflow_configs,
            creation_time=creation_time,
            tags = tags,
            read_option_configs = read_option_configs,
            staging_sink_write_option_configs = staging_sink_write_option_configs,
            online_sink_write_option_configs = online_sink_write_option_configs,
        )

    def _get_feature_transform_args(self, transform: dict):
        if transform.get("feature_transform_args") is None:
            return {}
        else:
            return transform["feature_transform_args"]

    def to_json(self) -> Dict:
        return {
            "name": self.name,
            "type": self.type,
            "data_type": self.data_type.value,
            "data_sources": self.data_sources,
            "owners": self.owners,
            "description": self.description,
            "start_time": str(self.start_time),
            "online": self.online,
            "offline": self.offline,
            "entity": self.entity,
            "feature_logic": self.feature_logic,
            "k8s_configs": self.k8s_configs,
            "processing_engine": self.processing_engine.value,
            "processing_engine_configs": self.processing_engine_configs.to_json(),
            "sink": self.sink,
            "creation_time": str(self.creation_time),
            "airflow_configs": self.airflow_configs,
            "created_at": str(self.creation_time),  # For backward compatibility
            "metadata" : self.tags,
            "read_option_configs": self.read_option_configs,
            "staging_sink_write_option_configs": self.staging_sink_write_option_configs,
            "online_sink_write_option_configs": self.online_sink_write_option_configs,
        }

    def to_register_json(self) -> Dict:
        return self.to_json()


import os
from typing import Optional, Type
from langchain_core.tools import BaseTool
from pydantic import BaseModel, Field
from langchain_core.runnables import RunnableConfig
from langchain_core.callbacks import CallbackManagerForToolRun

from gru.agents.checkpoint.task_results import TaskResultsRepository, TaskStatus
from gru.agents.task_server.celery import TaskMetadata, submit_task


class PythonCodeRunnerInput(BaseModel):
    repository: str = Field(description="code repository which has the code to execute")
    file_path_to_execute: str = Field(
        description="path of the python file in the repository"
    )
    arguments: list[str] = Field(
        description="arguments required to execute the python file"
    )
    run_id: str = Field(description="random unique id to identify each run")


class PythonCodeRunner(BaseTool):
    name: str = "run_python_code"
    description: str = (
        "use this tool to submit a python file present in a repository for execution."
    )
    args_schema: Type[BaseModel] = PythonCodeRunnerInput
    return_direct: bool = True

    class Config:
        extra = "allow"

    def __init__(self, token: str):
        task_results_repo = TaskResultsRepository()
        super().__init__(token=token, task_results_repo=task_results_repo)
        self.token = token
        self.task_results_repo = task_results_repo

    def _run(self, *args, **kwargs):
        return super()._run(*args, **kwargs)

    async def _arun(
        self,
        repository: str,
        file_path_to_execute: str,
        arguments: list[str],
        run_id: str,
        config: RunnableConfig,
        run_manager: Optional[CallbackManagerForToolRun] = None,
    ):
        agent_url = os.getenv("AGENT_URL")
        agent_name = os.getenv("AGENT_NAME")

        metadata = TaskMetadata(
            prompt_id=config["metadata"]["thread_id"],
            tool_call_id=run_id,
            agent_name=agent_name,
            agent_callback_url=f"http://{agent_url}/save-task-result",
        )

        task_attributes = {
            "repository": repository,
            "token": self.token,
            "file_path_to_execute": file_path_to_execute,
            "arguments": arguments,
        }

        submit_task(self.name, metadata, task_attributes)
        await self.task_results_repo.update(
            agent_name,
            config["metadata"]["thread_id"],
            self.name,
            run_id,
            TaskStatus.PROCESSING,
        )

        return "The python file has been submitted for execution with run_id {run_id}. Ask the user to check the status later."
